package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

public class TreeNode {
    private static final String ATTR_TYPE = "_type";
    private static final String ATTR_PREDICATE = "_predicate";
    private static final String ATTR_DESCRIPTION = "_description";
    private static final String ATTR_BASE_NODES = "_baseNodes";

    private UUID uuid;
    private final TreeDB db;
    protected String id;
    UUID parent;
    List<UUID> children = new ArrayList<UUID>();
    Map<String, Attribute> attributes = new LinkedHashMap<String, Attribute>();

    public TreeNode(TreeDB db, UUID uuid) {
        if (uuid == null)
            uuid = UUID.randomUUID();
        this.uuid = uuid;
        this.db = db;
    }

    public void setPredicate(String pred) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_PREDICATE, pred));
    }

    public String getPredicate() {
        if (getAttributeValue(ATTR_PREDICATE) == null) {
            return "";
        }
        return (String) getAttributeValue(ATTR_PREDICATE);
    }

    public void setType(String type) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_TYPE, type));
    }

    public final UUID getUUId() {
        return uuid;
    }

    public void setId(String id) {
        if (!this.id.equals(id)) {
            db.getStorage().changeNodeId(uuid, id);
        }
    }

    public String getId() {
        return id;
    }

    public String getType() {
        return (String) getAttributeValue(ATTR_TYPE);
    }

    public void setParent(TreeNode parent) {
        db.getStorage().moveNodeTo(uuid, parent.getUUId());
    }

    public UUID getParentUUId() {
        return parent;
    }

    public void removeChild(TreeNode node) {
        // TODO: find chield and remove it
        db.getStorage().remove(node.getUUId());
    }

    public void setDescription(String description) {
        attributes.put(ATTR_DESCRIPTION, new Attribute(AttributeType.STRING,
                ATTR_DESCRIPTION, description));
    }

    public String getDescription() {
        if (attributes.get(ATTR_DESCRIPTION) != null) {
            return (String) attributes.get(ATTR_DESCRIPTION).getValue();
        }
        return null;
    }

    public TreeDB getTreeDB() {
        return db;
    }

    void setUUId(UUID uuid) {
        this.uuid = uuid;
    }

    public TreeNode getParent() {
        if(parent==null) {
            // no parent
            assert this.getTreeDB().getRootNode().equals(this);
            return null;
        }
        return db.getNode(parent);
    }

    public int sizeChildren() {
        return children.size();
    }

    public TreeNode getChild(int i) {
        return db.getNode(children.get(i));
    }

    public String getQualifiedId() {
        if (getParent() == null) {
            return "/";
        }
        String parentQId = getTreeDB().getNode(parent).getQualifiedId();
        if (parentQId.equals("/")) {
            return parentQId + this.getId();
        }
        return parentQId + "/" + this.getId();
    }

    public TreeNode findChild(String id) {
        for (int i = 0; i < sizeChildren(); i++) {
            if (getChild(i).getId().equals(id)) {
                return getChild(i);
            }
        }
        return null;
    }

    @Override
    public String toString() {
        return getId() + "(" + getUUId() + ")";
    }

    public TreeNode createChild(String id) {
        if (findChild(id) != null)
            throw new IllegalArgumentException("child '" + id
                    + "' is already exists");
        if (getType().equals("DBRoot")) {
            return db.createChildNode(uuid, id, "Node");
        }
        return db.createChildNode(uuid, id, getType());
    }

    public TreeNode createChild(String id, String type) {
        return db.createChildNode(uuid, id, type);
    }

    public void delete() {
        getParent().removeChild(this);
    }

    public Attribute putAttribute(Attribute a) {
        Attribute result = attributes.get(a.getKey());
        attributes.put(a.getKey(), a);
        return result;
    }

    public Attribute removeAttribute(String key) {
        return attributes.remove(key);
    }

    public Attribute getAttribute(String key) {
        return attributes.get(key);
    }

    public Object getAttributeValue(String key) {
        return getAttributeValue(key, null);
    }

    public Object getAttributeValue(String key, Object def) {
        Attribute value = attributes.get(key);
        return value != null ? value.getValue() : def;
    }
    
    public int getIntAttributeValue(String key, int def) {
        Attribute value = attributes.get(key);
        return value != null ? (Integer)value.getValue() : def;
    }

    public int getIntAttributeValue(String key) {
        return getIntAttributeValue(key, 0);
    }

    public float getFloatAttributeValue(String key, float def) {
        Attribute value = attributes.get(key);
        return value != null ? (Float)value.getValue() : def;
    }

    public float getFloatAttributeValue(String key) {
        return getFloatAttributeValue(key, 0);
    }

    public boolean getBooleanAttributeValue(String key, boolean def) {
        Attribute value = attributes.get(key);
        return value != null ? (Boolean)value.getValue() : def;
    }

    public boolean getBooleanAttributeValue(String key) {
        return getBooleanAttributeValue(key, false);
    }

    public String getStringAttributeValue(String key, String def) {
        Attribute value = attributes.get(key);
        return value != null ? (String)value.getValue() : def;
    }

    public String getStringAttributeValue(String key) {
        return getStringAttributeValue(key, null);
    }

    @SuppressWarnings("unchecked")
    public List<String> getListAttributeValue(String key) {
        return (List<String>) getAttributeValue(key, null);
    }
    // TODO: resource (eclipse only now)
//    public IResource getResourceAttributeValue(String key) {
//        return (IResource) getAttributeValue(key, null);
//    }

    public Set<String> getAttributeKeys() {
        return attributes.keySet();
    }

    public void saveAttributes() {
        db.getStorage().saveAttributes(uuid, this.attributes.values());
    }

    @SuppressWarnings("unchecked")
    public void addBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null) {
            baseNodes = new ArrayList<String>();
        }
        baseNodes.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
    }

    @SuppressWarnings("unchecked")
    public boolean removeBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null || !baseNodes.remove(nodeQId)) {
            return false;
        }
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
        return true;
    }

    @SuppressWarnings("unchecked")
    public List<String> getBaseNodes() {
        if (getAttributeValue(ATTR_BASE_NODES) == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections
                .unmodifiableList((List<String>) getAttributeValue(ATTR_BASE_NODES));
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (super.equals(obj))
            return true;
        if (!(obj instanceof TreeNode))
            return false;
        TreeNode node = (TreeNode) obj;
        if (!node.getUUId().equals(uuid))
            return false;
        if (!getQualifiedId().equals(node.getQualifiedId()))
            return false;
        if (!getAttributeKeys().equals(node.getAttributeKeys()))
            return false;
        for (String key : getAttributeKeys()) {
            if (!getAttributeValue(key).equals(node.getAttributeValue(key)))
                return false;
        }
        return true;
    }
    public TreeNode[] getSortedChildren() {
        return getSortedChildren(null, null);
    }
    public TreeNode[] getSortedChildren(String type) {
        return getSortedChildren(null, type);
    }
    public TreeNode[] getSortedChildren(Comparator<TreeNode> c, String type) {
        ArrayList<TreeNode> result = new ArrayList<TreeNode>();
        for (int i = 0; i < sizeChildren(); i++) {
            TreeNode ch = getChild(i);
            if(type==null || ch.getType().equals(type)) {
                result.add(getChild(i));
            }
        }
        TreeNode[] r = (TreeNode[]) result.toArray(new TreeNode[0]);
        if (c != null) {
            Arrays.sort(r, c);
        } else {
            Arrays.sort(r, new Comparator<TreeNode>() {
                @Override
                public int compare(TreeNode o1, TreeNode o2) {
                    return o1.getId().compareTo(o2.getId());
                }
            });
        }
        return r;
    }
}
